// asmee.c

#define versionName "Emotion Engine assembler"
#include "asmx.h"

#define LIMIT 131071

enum
{
    o_Branch,
    o_Jump,
    o_OneReg,
    o_OneRegShift,
    o_TwoRegShift,
    o_TwoReg,
    o_TwoRegGap,
    o_ThreeReg,
    o_ThreeRegSwap,
    o_Serv,
    o_SYNCstype,
    o_Trap,
    o_TrapI,
    o_MFx,
    o_Imm,
    o_PREF,
    o_LUI,
    o_DataInterchange,
    o_Implied,      // NOP

//  o_Foo = o_LabelOp,
};

const char GPR[] = "ZERO AT V0 V1 A0 A1 A2 A3 T0 T1 T2 T3 T4 T5 T6 T7 S0 S1 S2 S3 S4 S5 S6 S7 T8 T9 K0 K1 GP SP FP RA";
//const char fregs[] = "FPR0 FPR1 FPR2 FPR3 FPR4 FPR5 FPR6 FPR7 FPR8 FPR9 FPR10 FPR11 FPR12 FPR13 FPR14 FPR15 FPR16 FPR17 FPR18 FPR19 FPR20 FPR21 FPR22 FPR23 FPR24 FPR25 FPR26 FPR27 FPR28 FPR29 FPR30 FPR31";
//const char sregs[] = "";
//const char other[] = "PC";

struct OpcdRec EE_opcdTab[] =
{
    {"BEQ",     o_Branch,          0x10000000},
    {"BNE",     o_Branch,          0x14000000},
    
    {"BEQL",    o_Branch,          0x50000000},
    {"BNEL",    o_Branch,          0x54000000},

    {"BL",      o_Branch,          0x48000001},
    
    {"TEQ",     o_Trap,            0x00000034},
    {"TGE",     o_Trap,            0x00000030},
    {"TGEU",    o_Trap,            0x00000031},
    {"TLT",     o_Trap,            0x00000032},
    {"TLTU",    o_Trap,            0x00000033},
    {"TNE",     o_Trap,            0x00000036},
    
    {"TEQI",    o_TrapI,           0x040C0000},
    {"TGEI",    o_TrapI,           0x04080000},
    {"TGEIU",   o_TrapI,           0x04090000},
    {"TLTI",    o_TrapI,           0x040A0000},
    {"TLTIU",   o_TrapI,           0x040B0000},
    {"TNEI",    o_TrapI,           0x040E0000},
    
    {"J",       o_Jump,            0x08000000},
    {"JAL",     o_Jump,            0x0C000000},

    {"JR",      o_OneReg,          0x00000008},
    {"MTHI",    o_OneReg,          0x00000011},
    {"MTLO",    o_OneReg,          0x00000013},

    {"BGEZ",    o_OneRegShift,     0x04010000},
    {"BGEZAL",  o_OneRegShift,     0x04110000},
    {"BGEZALL", o_OneRegShift,     0x04130000},
    {"BGEZL",   o_OneRegShift,     0x04030000},
    {"BGTZ",    o_OneRegShift,     0x1C000000},
    {"BGTZL",   o_OneRegShift,     0x5C000000},
    {"BLEZ",    o_OneRegShift,     0x18000000},
    {"BLEZL",   o_OneRegShift,     0x58000000},
    {"BLTZ",    o_OneRegShift,     0x04000000},
    {"BLTZAL",  o_OneRegShift,     0x04100000},
    {"BLTZALL", o_OneRegShift,     0x04120000},
    {"BLTZL",   o_OneRegShift,     0x04020000},

    {"DSLL",    o_TwoRegShift,     0x00000038},
    {"DSLL32",  o_TwoRegShift,     0x0000003C},
    {"DSRA",    o_TwoRegShift,     0x0000003B},
    {"DSRA32",  o_TwoRegShift,     0x0000003F},
    {"DSRL",    o_TwoRegShift,     0x0000003A},
    {"DSRL32",  o_TwoRegShift,     0x0000003E},
    {"SLL",     o_TwoRegShift,     0x00000000},
    {"SRA",     o_TwoRegShift,     0x00000003},
    {"SRL",     o_TwoRegShift,     0x00000002},
    
    {"DIV",     o_TwoReg,          0x0000001A},
    {"DIVU",    o_TwoReg,          0x0000001B},
    {"MULT",    o_TwoReg,          0x00000018},
    {"MULTU",   o_TwoReg,          0x00000019},
    
    {"JALR",    o_TwoRegGap,       0x00000009},
    
    {"DSLLV",   o_ThreeReg,        0x00000014},
    {"DSRAV",   o_ThreeReg,        0x00000017},
    {"DSRLV",   o_ThreeReg,        0x00000016},
    {"SLLV",    o_ThreeReg,        0x00000004},
    {"SRAV",    o_ThreeReg,        0x00000007},
    {"SRLV",    o_ThreeReg,        0x00000006},
    
    {"ADD",     o_ThreeRegSwap,    0x00000020},
    {"ADDU",    o_ThreeRegSwap,    0x00000021},
    {"AND",     o_ThreeRegSwap,    0x00000024},
    {"DADD",    o_ThreeRegSwap,    0x0000002C},
    {"DADDU",   o_ThreeRegSwap,    0x0000002D},
    {"DSUB",    o_ThreeRegSwap,    0x0000002E},
    {"DSUBU",   o_ThreeRegSwap,    0x0000002F},
    {"MOVN",    o_ThreeRegSwap,    0x0000000B},
    {"MOVZ",    o_ThreeRegSwap,    0x0000000A},
    {"NOR",     o_ThreeRegSwap,    0x00000027},
    {"OR",      o_ThreeRegSwap,    0x00000025},
    {"SLT",     o_ThreeRegSwap,    0x0000002A},
    {"SLTU",    o_ThreeRegSwap,    0x0000002B},
    {"SUB",     o_ThreeRegSwap,    0x00000022},
    {"SUBU",    o_ThreeRegSwap,    0x00000023},
    {"XOR",     o_ThreeRegSwap,    0x00000026},

    {"BREAK",   o_Serv,            0x0000000D},
    {"SYSCALL", o_Serv,            0x0000000C},
    
    {"SYNC",    o_SYNCstype,       0x0000000F},
    {"SYNC.L",  o_SYNCstype,       0x0000000F},
    {"SYNC.P",  o_SYNCstype,       0x0000040F},
    
    {"ADDI",    o_Imm,             0x20000000},
    {"ADDIU",   o_Imm,             0x24000000},
    {"ANDI",    o_Imm,             0x30000000},
    {"DADDI",   o_Imm,             0x60000000},
    {"DADDIU",  o_Imm,             0x64000000},
    {"ORI",     o_Imm,             0x34000000},
    {"SLTI",    o_Imm,             0x28000000},
    {"SLTIU",   o_Imm,             0x2C000000},
    {"XORI",    o_Imm,             0x38000000},
    
    {"LB",      o_DataInterchange, 0x80000000},
    {"LBU",     o_DataInterchange, 0x90000000},
    {"LD",      o_DataInterchange, 0xDC000000},
    {"LDL",     o_DataInterchange, 0x68000000},
    {"LDR",     o_DataInterchange, 0x6C000000},
    {"LH",      o_DataInterchange, 0x84000000},
    {"LHU",     o_DataInterchange, 0x94000000},
    {"LW",      o_DataInterchange, 0x8C000000},
    {"LWL",     o_DataInterchange, 0x88000000},
    {"LWR",     o_DataInterchange, 0x98000000},
    {"LWU",     o_DataInterchange, 0x9C000000},
    {"SB",      o_DataInterchange, 0xA0000000},
    {"SD",      o_DataInterchange, 0xFC000000},
    {"SDL",     o_DataInterchange, 0xB0000000},
    {"SDR",     o_DataInterchange, 0xB4000000},
    {"SH",      o_DataInterchange, 0xA4000000},
    {"SW",      o_DataInterchange, 0xAC000000},
    {"SWL",     o_DataInterchange, 0xA8000000},
    {"SWR",     o_DataInterchange, 0xB8000000},
    
    {"LUI",     o_LUI,             0x3C000000},
    
    {"PREF",    o_PREF,            0xCC000000},

    {"MFHI",    o_MFx,             0x00000010},
    {"MFLO",    o_MFx,             0x00000012},

    {"NOP",     o_Implied,         0xE1A00000}, // pseudo-instruction = MOV R0,R0

    {"",        o_Illegal,         0}
};


// --------------------------------------------------------------


int EEvalBranch()
{
    long val;

    val = Eval();
    val = val - locPtr - 4;
    if (!errFlag && ((val & 3) || val < ~LIMIT || val > LIMIT))
        Error("Long branch out of range");

    return val;
}


int EE_DoCPUOpcode(int typ, int parm)
{
    int     val1, val2;
    long    val;
    int     i;
    Str255  word;
    int     token;
    int     reg1,reg2,reg3;

    // get condition from opcode wildcard
    word[0]=0;
    if (isalnum(*linePtr))
        GetWord(word);

    switch(typ)
    {
        case o_Branch:
            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(GPR);
            if (CheckReg(reg2)) break;

            if (Comma()) break;

            val = EEvalBranch(18);
            InstrL(parm | (reg1 << 21) | (reg2 << 16) | ((val >> 2) & 0xFFFF));
            break;

        case o_Jump:

            val = Eval();

            InstrL(parm | ((val & 0xFFFFFFF) >> 2));
            break;

        case o_OneReg:

            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            InstrL(parm | (reg1 << 21));
            break;

        case o_OneRegShift:

            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            val = EEvalBranch(18);
            InstrL(parm | (reg1 << 21) | ((val >> 2) & 0xFFFF));
            break;

        case o_TwoReg:

            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(GPR);
            if (CheckReg(reg2)) break;

            InstrL(parm | (reg1 << 21) | (reg2 << 16));
            break;

        case o_TwoRegGap:

            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            token = GetWord(word);
            if (token == ',')
            {
             reg2 = GetReg(GPR);
             if (CheckReg(reg2)) break;
            }
            else
            {
             reg2=reg1;
             reg1=31;
            }

            InstrL(parm | (reg1 << 21) | (reg2 << 11));
            break;

        case o_TwoRegShift:

            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(GPR);
            if (CheckReg(reg2)) break;

            if (Comma()) break;

            val1 = Eval();

            InstrL(parm | (reg2 << 16) | (reg1 << 11) | ((val1 & 0x1F) << 6));
            break;

        case o_ThreeReg:

            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(GPR);
            if (CheckReg(reg2)) break;

            if (Comma()) break;

            reg3 = GetReg(GPR);
            if (CheckReg(reg3)) break;

            InstrL(parm | (reg3 << 21) | (reg2 << 16) | (reg1 << 11));
            break;

        case o_ThreeRegSwap:

            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(GPR);
            if (CheckReg(reg2)) break;

            if (Comma()) break;

            reg3 = GetReg(GPR);
            if (CheckReg(reg3)) break;

            InstrL(parm | (reg2 << 21) | (reg3 << 16) | (reg1 << 11));
            break;

        case o_Imm:
            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(GPR);
            if (CheckReg(reg2)) break;

            if (Comma()) break;

            val1 = Eval();

            InstrL(parm | (reg2 << 21) | (reg1 << 16) | val1);
            break;

        case o_Serv:
            val = Eval();

            InstrL(parm  | ((val << 6) & 0x3FFFFC0));
            break;
            
        case o_DataInterchange:
            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            val1 = Eval();
            token = GetWord(word);
            if (token == '(')
            {
             reg2 = GetReg(GPR);
             if (CheckReg(reg2)) break;
             if (RParen()) break;
            }

            InstrL(parm | (reg2 << 21) | (reg1 << 16) | val1);
            break;

        case o_LUI:
            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            val1 = Eval();

            InstrL(parm | (reg1 << 16) | val1);
            break;

        case o_MFx:

            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            InstrL(parm | (reg1 << 11));
            break;
            
        case o_PREF:
            val1 = Eval();

            if (Comma()) break;

            val2 = Eval();
            token = GetWord(word);
            if (token == '(')
            {
             reg1 = GetReg(GPR);
             if (CheckReg(reg1)) break;
             if (RParen()) break;
            }

            InstrL(parm | (reg1 << 21) | ((val1 & 31) << 16) | val2);
            break;

        case o_SYNCstype:
            val1 = Eval();

            InstrL(parm  | ((val1 & 0xF) << 6));
            break;

        case o_Trap:
            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            reg2 = GetReg(GPR);
            if (CheckReg(reg2)) break;

            token = GetWord(word);
            if (token == ',') val1 = Eval();
            else val1=0;

            InstrL(parm | (reg1 << 21) | (reg2 << 16) | (val1 & 0x3FF) << 6);
            break;
            
        case o_TrapI:
            reg1 = GetReg(GPR);
            if (CheckReg(reg1)) break;

            if (Comma()) break;

            val1 = Eval();

            InstrL(parm | (reg1 << 21) | val1);
            break;

if (pass == 2)
{
    printf("*** %.8X *** %s\n",val1,line);
}

//            InstrL(   (cond << 28) | (val1 & 0x0FFF));
            break;

        case o_Implied:      // NOP
            InstrL(parm);
            break;

        default:
            return 0;
            break;
    }

    if (locPtr & 3)
    {
        Error("Code at non-longword-aligned address");
        // deposit extra bytes to reset alignment and prevent further errors
        for (i=locPtr & 3; i<4; i++)
            InstrAddB(0);
        // note: Inserting bytes in front won't work because offsets have already been assembled.
        // The line could be re-assembled by recursively calling DoCPUOpcode, but then
        // the label before the instruction would still be at the odd address.
    }

    return 1;
}


void AsmEEInit(void)
{
    char *p;

    p = AddAsm(versionName, &EE_DoCPUOpcode, NULL, NULL);
    AddCPU(p, "R5900_BE", 0, BIG_END,    ADDR_24, LIST_24, 8, 0, EE_opcdTab);
    AddCPU(p, "R5900_LE", 0, LITTLE_END, ADDR_24, LIST_24, 8, 0, EE_opcdTab);
}
